package cn.com.qmth.examcloud.core.oe.task.service.impl;

import cn.com.qmth.examcloud.commons.exception.StatusException;
import cn.com.qmth.examcloud.core.oe.student.dao.ExamCaptureQueueRepo;
import cn.com.qmth.examcloud.core.oe.student.dao.ExamCaptureRepo;
import cn.com.qmth.examcloud.core.oe.student.dao.entity.ExamCaptureEntity;
import cn.com.qmth.examcloud.core.oe.task.service.ExamCaptureQueueService;
import cn.com.qmth.examcloud.core.oe.task.service.ExamCaptureService;
import cn.com.qmth.examcloud.core.oe.task.service.ExamRecordDataService;
import cn.com.qmth.examcloud.core.oe.task.service.bean.CalculateFaceCheckResultInfo;
import cn.com.qmth.examcloud.core.oe.task.service.bean.ExamCaptureQueueInfo;
import cn.com.qmth.examcloud.support.Constants;
import cn.com.qmth.examcloud.support.cache.CacheHelper;
import cn.com.qmth.examcloud.support.cache.bean.SysPropertyCacheBean;
import cn.com.qmth.examcloud.support.enums.ExamProperties;
import cn.com.qmth.examcloud.support.examing.ExamRecordData;
import cn.com.qmth.examcloud.support.helper.ExamCacheTransferHelper;
import cn.com.qmth.examcloud.support.helper.FaceBiopsyHelper;
import cn.com.qmth.examcloud.web.redis.RedisClient;
import org.apache.commons.lang3.StringUtils;
import org.json.JSONException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.Example;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import java.math.BigDecimal;
import java.util.List;

/**
 * @author chenken
 * @date 2018年9月5日 下午3:31:37
 * @company QMTH
 * @description 考试抓拍服务实现
 */
@SuppressWarnings("ALL")
@Service("examCaptureService")
public class ExamCaptureServiceImpl implements ExamCaptureService {

    private static final Logger log = LoggerFactory.getLogger(ExamCaptureServiceImpl.class);

    @Autowired
    private ExamCaptureRepo examCaptureRepo;

    @Autowired
    private ExamCaptureQueueRepo examCaptureQueueRepo;

    @Autowired
    private ExamCaptureQueueService examCaptureQueueService;

    @Autowired
    private ExamRecordDataService examRecordDataService;

    public static final String TEMP_FILE_EXP = "face_compare/";

    @Autowired
    private RedisClient redisClient;

    /**
     * 保存人脸检测最终结果并删除队列
     *
     * @param examCapture      抓拍照片最终检测最终结果实体
     * @param examCaptureQueue 抓拍照片队列表
     */
    @Override
    @Transactional
    public void saveExamCaptureAndDeleteQueue(ExamCaptureQueueInfo examCaptureQueue) {
        ExamCaptureEntity examCapture = getExamCaptureFromQueue(examCaptureQueue);

        //同一考试记录下如果有重复的照片，则直接跳过
        ExamCaptureEntity query = new ExamCaptureEntity();
        query.setExamRecordDataId(examCapture.getExamRecordDataId());
        query.setFileName(examCapture.getFileName());
        Example<ExamCaptureEntity> example = Example.of(query);
        //照片处理结果中如果已存在，则以已有的数据为准
        if (!examCaptureRepo.exists(example)) {
            examCaptureRepo.save(examCapture);
        }
        //删除队列中的记录
        examCaptureQueueRepo.deleteById(examCaptureQueue.getId());
    }

    /**
     * 计算人脸检测结果
     * 相片数=0,系统判断为违纪,自动审核
     * 考试记录为异常逻辑(进入待审)：
     * 1.陌生人次数>0
     * 2.face++阈值 = 0 && 百度真实性阈值 > 0
     * 真实性百分比<百度真实性阈值
     * 3.face++阈值 > 0 && 百度真实性阈值 = 0
     * face++成功率<face++阈值
     * 4.face++阈值 > 0 && 百度真实性阈值 > 0
     * face++成功率<face++阈值 ||
     * 真实性百分比<百度真实性阈值
     *
     * @param examRecordData
     * @return
     */
    @Override
    public CalculateFaceCheckResultInfo calculateFaceCheckResult(Long examRecordDataId) {
        ExamRecordData examRecordData = examRecordDataService.getExamRecordDataCache(examRecordDataId);

        Long rootOrgId = examRecordData.getRootOrgId();
        Long examId = examRecordData.getExamId();
        Long orgId = examRecordData.getOrgId();
        Long studentId = examRecordData.getStudentId();

        //人脸最终检测结果
        CalculateFaceCheckResultInfo resultInfo = new CalculateFaceCheckResultInfo();

        //未开启人脸检测,直接认为正常数据
        if (!FaceBiopsyHelper.isFaceEnable(rootOrgId, examId, studentId)) {
            resultInfo.setIsWarn(false);
            return resultInfo;
        }

        List<ExamCaptureEntity> examCaptureList = examCaptureRepo.findByExamRecordDataId(examRecordData.getId());

        //无照片违纪
        if (examCaptureList == null || examCaptureList.size() == 0) {
            resultInfo.setIsWarn(true);//有异常
            resultInfo.setIsIllegality(true);//违纪
            resultInfo.setNoPhotoAndIllegality(true);//无照片违纪
            return resultInfo;
        }

        //根据照片结果计算
        return calculateByCaptureResult(examCaptureList, examRecordData.getExamId(), examRecordData.getStudentId());
    }

    /**
     * 计算人脸检测数据
     * 陌生人记录数、成功次数、失败次数、成功率
     *
     * @param examCaptureEntityList
     * @param examId
     * @param studentId
     * @return
     */
    private CalculateFaceCheckResultInfo calculateByCaptureResult(List<ExamCaptureEntity> examCaptureEntityList,
                                                                  Long examId, Long studentId) {
        int strangerCount = 0;    // 人脸比较陌生人记录数
        int succCount = 0;        // 人脸比较成功次数
        int falseCount = 0;        // 人脸比较失败次数
        double succPercent = 0d;    // 人脸比较成功率
        int livenessSuccessCount = 0;//百度活体检测成功次数
        double livenessSuccessPercent = 0D;//百度活体检测成功率
        for (ExamCaptureEntity examCaptureEntity : examCaptureEntityList) {
            if (examCaptureEntity.getIsPass() != null && examCaptureEntity.getIsPass()) {
                succCount++;
            } else {
                falseCount++;
            }
            if (examCaptureEntity.getIsStranger() != null && examCaptureEntity.getIsStranger()) {
                strangerCount++;
            }
            livenessSuccessCount += calculateFacelivenessSuccessCount(examCaptureEntity.getFacelivenessResult());
        }
        int allNum = examCaptureEntityList.size();
        BigDecimal bg = new BigDecimal(((double) succCount / allNum) * 100);
        succPercent = bg.setScale(2, BigDecimal.ROUND_HALF_UP).doubleValue();// 人脸比较成功率

        //活体检测最终结果
        CalculateFaceCheckResultInfo resultInfo = new CalculateFaceCheckResultInfo();

        resultInfo.setFaceTotalCount(allNum);//检测总次数
        resultInfo.setFaceSuccessPercent(succPercent);//成功率
        resultInfo.setFaceStrangerCount(strangerCount);//有陌生人的次数
        resultInfo.setFaceSuccessCount(succCount);//成功次数
        resultInfo.setFaceFailedCount(falseCount);//失败次数

        //计算百度活体检测通过率
        BigDecimal livenessSuccessBg = new BigDecimal(((double) livenessSuccessCount / allNum) * 100);
        livenessSuccessPercent = livenessSuccessBg.setScale(2, BigDecimal.ROUND_HALF_UP).doubleValue();//百度活体检测成功率
        resultInfo.setBaiduFaceLivenessSuccessPercent(livenessSuccessPercent);

        //陌生人个数>0
        if (resultInfo.getFaceStrangerCount() > 0) {
            resultInfo.setIsWarn(true);
            return resultInfo;
        }

        //人脸识别阀值
        String warnThresholdStr = ExamCacheTransferHelper.getCachedExamProperty(examId,
                studentId, ExamProperties.WARN_THRESHOLD.name()).getValue();
        if (StringUtils.isBlank(warnThresholdStr)) {
            throw new StatusException("400101", "人脸检测预警阈值未设置");
        }

        //人脸真实性(百度活体检测)通过阀值
        String liveWarnThresholdStr = ExamCacheTransferHelper.getCachedExamProperty(examId,
                studentId, ExamProperties.LIVING_WARN_THRESHOLD.name()).getValue();

        if (StringUtils.isBlank(liveWarnThresholdStr)) {
            throw new StatusException("400102", "人脸真实性阈值未设置");
        }

        double warnThreshold = Double.parseDouble(warnThresholdStr);
        double livenessThreshold = Double.parseDouble(liveWarnThresholdStr);

        if (warnThreshold == 0d && livenessThreshold > 0d) {
            resultInfo.setIsWarn(livenessSuccessPercent < livenessThreshold);
        } else if (warnThreshold > 0d && livenessThreshold == 0d) {
            resultInfo.setIsWarn(succPercent < warnThreshold);
        } else if (warnThreshold > 0d && livenessThreshold > 0d) {
            resultInfo.setIsWarn(succPercent < warnThreshold || livenessSuccessPercent < livenessThreshold);
        }

        return resultInfo;
    }

    /**
     * 计算百度活体检测成功数量
     */
    private int calculateFacelivenessSuccessCount(String facelivenessResult) {
        if (StringUtils.isNotBlank(facelivenessResult)) {
            // 兼容C端学生端情况，人脸真实性值只存在0或1两种值
            if ("0".equals(facelivenessResult)) {
                return 0;
            }
            if ("1".equals(facelivenessResult)) {
                return 1;
            }

            org.json.JSONObject jsonObject;
            try {
                jsonObject = new org.json.JSONObject(facelivenessResult);
                if (jsonObject.has("error_code") && jsonObject.getInt("error_code") == 0 && jsonObject.has("result")) {
                    org.json.JSONObject resultJson = jsonObject.getJSONObject("result");
                    if (resultJson.has("face_liveness")) {
                        double faceLivenessVal = resultJson.getDouble("face_liveness");

                        Double baiduFacelivenessThreshold;
                        SysPropertyCacheBean baiduFacelivenessThresholdProperty = CacheHelper.getSysProperty("$baidu.faceliveness.threshold");
                        if (!baiduFacelivenessThresholdProperty.getHasValue()) {
                            baiduFacelivenessThreshold = Constants.DEFAULT_BAIDU_FACELIVENESS_THRESHOLD;
                        } else {
                            baiduFacelivenessThreshold = Double.valueOf(baiduFacelivenessThresholdProperty.getValue().toString());
                        }

                        if (faceLivenessVal > baiduFacelivenessThreshold) {
                            return 1;
                        }
                    }
                }
            } catch (JSONException e) {
                log.error(e.getMessage(), e);
                return 0;
            }
        }
        return 0;
    }

    private ExamCaptureEntity getExamCaptureFromQueue(ExamCaptureQueueInfo queue) {
        long currentTimeMillis = System.currentTimeMillis();
        long createTimeMillis = queue.getCreationTime().getTime();
        long faceCompareStartTimeMillis = queue.getFaceCompareStartTime();

        ExamCaptureEntity resultEntity = new ExamCaptureEntity();

        if (StringUtils.length(queue.getCameraInfos()) >= Constants.VM_CAMERA_SIZE_LIMIT) {
            resultEntity.setCameraInfos(Constants.VM_CAMERA_WARN);
            log.warn("虚拟摄像头信息超长！ " + queue.getExamRecordDataId());
        } else {
            resultEntity.setCameraInfos(queue.getCameraInfos());
        }

        resultEntity.setExamRecordDataId(queue.getExamRecordDataId());
        resultEntity.setExtMsg(queue.getExtMsg());
        resultEntity.setFaceCompareResult(queue.getFaceCompareResult());
        resultEntity.setFacelivenessResult(queue.getFacelivenessResult());
        resultEntity.setFileName(queue.getFileName());
        resultEntity.setFileUrl(queue.getFileUrl());
        resultEntity.setHasVirtualCamera(queue.getHasVirtualCamera());
        resultEntity.setIsStranger(queue.getStranger());
        resultEntity.setIsPass(queue.getPass());
        resultEntity.setProcessTime(currentTimeMillis - createTimeMillis);//从进队列到处理完毕的时间
        resultEntity.setUsedTime(currentTimeMillis - faceCompareStartTimeMillis);//从开始处理到处理完毕的时间
        return resultEntity;
    }

}
