package cn.com.qmth.examcloud.core.oe.task.service.impl;

import cn.com.qmth.examcloud.commons.exception.StatusException;
import cn.com.qmth.examcloud.core.oe.student.api.ExamRecordDataCloudService;
import cn.com.qmth.examcloud.core.oe.student.api.request.CalcExamScoreReq;
import cn.com.qmth.examcloud.core.oe.student.api.request.CalcFaceBiopsyResultReq;
import cn.com.qmth.examcloud.core.oe.student.api.response.CalcExamScoreResp;
import cn.com.qmth.examcloud.core.oe.student.api.response.CalcFaceBiopsyResultResp;
import cn.com.qmth.examcloud.core.oe.student.dao.ExamCaptureQueueRepo;
import cn.com.qmth.examcloud.core.oe.student.dao.ExamRecordDataRepo;
import cn.com.qmth.examcloud.core.oe.student.dao.entity.ExamRecordDataEntity;
import cn.com.qmth.examcloud.core.oe.task.service.ExamCaptureQueueService;
import cn.com.qmth.examcloud.core.oe.task.service.ExamCaptureService;
import cn.com.qmth.examcloud.core.oe.task.service.ExamRecordDataService;
import cn.com.qmth.examcloud.core.oe.task.service.bean.CalculateFaceCheckResultInfo;
import cn.com.qmth.examcloud.support.Constants;
import cn.com.qmth.examcloud.support.enums.ExamRecordStatus;
import cn.com.qmth.examcloud.support.enums.HandInExamType;
import cn.com.qmth.examcloud.support.enums.IsSuccess;
import cn.com.qmth.examcloud.support.examing.ExamRecordData;
import cn.com.qmth.examcloud.support.helper.FaceBiopsyHelper;
import cn.com.qmth.examcloud.support.redis.RedisKeyHelper;
import cn.com.qmth.examcloud.web.exception.SequenceLockException;
import cn.com.qmth.examcloud.web.helpers.SequenceLockHelper;
import cn.com.qmth.examcloud.web.redis.RedisClient;
import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.jdbc.core.BeanPropertyRowMapper;
import org.springframework.jdbc.core.JdbcTemplate;
import org.springframework.stereotype.Service;

import java.util.Arrays;
import java.util.Date;
import java.util.List;
import java.util.stream.Collectors;

/**
 * @Description 考试记录服务
 * @Author lideyin
 * @Date 2019/12/18 14:37
 * @Version 1.0
 */
@Service("examRecordDataService")
public class ExamRecordDataServiceImpl implements ExamRecordDataService {

    private static final Logger log = LoggerFactory.getLogger(ExamRecordDataServiceImpl.class);

    @Autowired
    private RedisClient redisClient;

    @Autowired
    private ExamCaptureQueueRepo examCaptureQueueRepo;

    @Autowired
    private ExamCaptureQueueService examCaptureQueueService;

    @Autowired
    private ExamRecordDataRepo examRecordDataRepo;

    @Autowired
    private ExamCaptureService examCaptureService;

    @Autowired
    private ExamRecordDataCloudService examRecordDataCloudService;

    @Autowired
    private JdbcTemplate jdbcTemplate;

    @Override
    public void saveExamRecordDataCache(Long examRecordDataId, ExamRecordData data) {
        String key = RedisKeyHelper.getBuilder().examRecordDataKey(examRecordDataId);
        redisClient.set(key, data, 2592000);
    }

    @Override
    public ExamRecordData getExamRecordDataCache(Long examRecordDataId) {
        String key = RedisKeyHelper.getBuilder().examRecordDataKey(examRecordDataId);
        return redisClient.get(key, ExamRecordData.class);
    }

    @Override
    public void deleteExamRecordDataCache(Long examRecordDataId) {
        String key = RedisKeyHelper.getBuilder().examRecordDataKey(examRecordDataId);
        redisClient.delete(key);
    }

    /**
     * 计算自动审核结果
     *
     * @param isNoPhotoAndIllegality
     * @param faceVerifyResult
     * @param isIllegality
     * @return
     */
    @Override
    public Boolean calcAutoAuditResult(Boolean isNoPhotoAndIllegality, IsSuccess faceVerifyResult, Boolean isIllegality) {
        //无照片违纪自动审核
        if (isNoPhotoAndIllegality != null && isNoPhotoAndIllegality) {
            return true;
        }

        //活体检测失败违纪自动审核
        if (null != faceVerifyResult && IsSuccess.FAILED == faceVerifyResult && null != isIllegality && isIllegality) {
            return true;
        }

        return null;
    }

    /**
     * 交卷后续处理
     */
    @Override
    public ExamRecordData processAfterHandInExam(Long examRecordDataId) {
        ExamRecordData examRecordData = getExamRecordDataCache(examRecordDataId);
        return this.processAfterHandInExam(examRecordData);
    }

    @Override
    public ExamRecordData processAfterHandInExam(ExamRecordData examRecordData) {
        //只有开启的人脸的才会进行如下处理
        if (FaceBiopsyHelper.isFaceEnable(examRecordData.getRootOrgId(), examRecordData.getExamId(), examRecordData.getStudentId())) {

            // 判断是否存在未处理的抓拍照片比对
            int queueSize = examCaptureQueueRepo.countByExamRecordDataId(examRecordData.getId());
            if (queueSize > 0) {
                final String lockKey = Constants.FACE_COMPARE_LOCK_PREFIX + examRecordData.getId();
                try {
                    SequenceLockHelper.getLockSimple(lockKey);

                    // 处理剩余未比对的抓拍照片
                    examCaptureQueueService.handlerExamCaptureQueuesByExamRecordDataId(examRecordData.getId());
                } catch (Exception e) {
                    if (e instanceof SequenceLockException) {
                        // 若锁问题，下次会继续执行
                        log.warn("当前人脸比对获取锁失败！examRecordDataId:{}, redisKey:{}", examRecordData.getId(), lockKey);
                    } else {
                        // 若异常，下次会继续执行（需要排查原因）
                        log.error("当前人脸比对处理失败！examRecordDataId:{}, error:{}", examRecordData.getId(), e.getMessage());
                    }

                    throw new StatusException(Constants.CAPTURE_PROCESSING_STATUS_CODE, "抓拍照片比对尚未处理完！");
                } finally {
                    SequenceLockHelper.releaseLockSimple(lockKey);
                }
            }

            // 计算人脸检测结果
            CalculateFaceCheckResultInfo faceCheckResult = examCaptureService.calculateFaceCheckResult(examRecordData.getId());
            if (null != faceCheckResult.getFaceTotalCount()) {
                examRecordData.setFaceTotalCount(faceCheckResult.getFaceTotalCount());
            }
            if (null != faceCheckResult.getFaceSuccessPercent()) {
                examRecordData.setFaceSuccessPercent(faceCheckResult.getFaceSuccessPercent());
            }
            if (null != faceCheckResult.getFaceStrangerCount()) {
                examRecordData.setFaceStrangerCount(faceCheckResult.getFaceStrangerCount());
            }
            if (null != faceCheckResult.getFaceSuccessCount()) {
                examRecordData.setFaceSuccessCount(faceCheckResult.getFaceSuccessCount());
            }
            if (null != faceCheckResult.getFaceFailedCount()) {
                examRecordData.setFaceFailedCount(faceCheckResult.getFaceFailedCount());
            }
            if (null != faceCheckResult.getBaiduFaceLivenessSuccessPercent()) {
                examRecordData.setBaiduFaceLivenessSuccessPercent(faceCheckResult.getBaiduFaceLivenessSuccessPercent());
            }

            // 计算活体检测结果
            CalcFaceBiopsyResultReq req = new CalcFaceBiopsyResultReq();
            req.setExamRecordDataId(examRecordData.getId());
            CalcFaceBiopsyResultResp calcFaceBiopsyResultResp = examRecordDataCloudService.calcFaceBiopsyResult(req);
            if (null != calcFaceBiopsyResultResp.getFaceVerifyResult()) {
                examRecordData.setFaceVerifyResult(calcFaceBiopsyResultResp.getFaceVerifyResult());
            }

            //判断是否违纪
            if (null != calcFaceBiopsyResultResp.getFaceVerifyResult() &&
                    IsSuccess.FAILED == calcFaceBiopsyResultResp.getFaceVerifyResult()) {
                examRecordData.setIsIllegality(true);//活检失败，直接判定为违纪
            } else {
                //人脸检测或活检中有一个判断为违纪，则为违纪
                boolean bi = (null != faceCheckResult.getIsIllegality() && faceCheckResult.getIsIllegality()) ||
                        (null != calcFaceBiopsyResultResp.getIsIllegality() && calcFaceBiopsyResultResp.getIsIllegality());
                examRecordData.setIsIllegality(bi);
            }

            //判断数据是否异常
            boolean bw = (null != faceCheckResult.getIsWarn() && faceCheckResult.getIsWarn()) ||
                    (null != calcFaceBiopsyResultResp.getIsWarn() && calcFaceBiopsyResultResp.getIsWarn());
            examRecordData.setIsWarn(bw);//人脸检测或活检中有一个判断为异常，则为异常数据

            // 违纪自动审核
            Boolean isAudit = this.calcAutoAuditResult(faceCheckResult.getNoPhotoAndIllegality(),
                    calcFaceBiopsyResultResp.getFaceVerifyResult(), calcFaceBiopsyResultResp.getIsIllegality());
            if (null != isAudit) {
                examRecordData.setIsAudit(isAudit);
            }
        }

        //自动计算客观分
        CalcExamScoreReq cesReq = new CalcExamScoreReq();
        cesReq.setExamRecordDataId(examRecordData.getId());
        CalcExamScoreResp calcExamScoreResp = examRecordDataCloudService.calcExamScore(cesReq);
        examRecordData.setObjectiveScore(calcExamScoreResp.getObjectiveScore());
        examRecordData.setObjectiveAccuracy(calcExamScoreResp.getObjectiveAccuracy());
        examRecordData.setSuccPercent(calcExamScoreResp.getSuccPercent());
        examRecordData.setTotalScore(calcExamScoreResp.getTotalScore());

        // 更新考试状态
        examRecordData.setExamRecordStatus(examRecordData.getHandInExamType() == HandInExamType.MANUAL ?
                ExamRecordStatus.EXAM_END : ExamRecordStatus.EXAM_OVERDUE);

        //更新考试记录中的相关数据
        this.saveExamRecordDataCache(examRecordData.getId(), examRecordData);

        // 更新考试记录状态
        examRecordDataRepo.updateExamRecordStatusById(examRecordData.getExamRecordStatus(), new Date(), examRecordData.getId());

        return examRecordData;
    }

    @Override
    public boolean isVirtualToWaiting(Long examId) {
        String valueSql = "SELECT value FROM ec_e_exam_prop WHERE exam_id =" + examId + " AND key_id=45";
        List<Boolean> list = jdbcTemplate.query(valueSql, (rs, rowNum) -> rs.getBoolean("value"));
        return !CollectionUtils.isEmpty(list) && list.get(0);
    }

    /**
     * 按ID分片获取未同步的考试记录列表
     */
    @Override
    public List<ExamRecordDataEntity> getExamRecordDataUnSyncList(int shardTotal, int shardIndex, int batchSize, long startId, ExamRecordStatus... status) {
        StringBuilder sql = new StringBuilder()
                .append(" select id,root_org_id,student_id,exam_record_status")
                .append(" from ec_oes_exam_record_data")
                .append(" where sync_status='UNSYNC'");

        if (ArrayUtils.isNotEmpty(status)) {
            List<String> values = Arrays.stream(status).map(e -> "'" + e.name() + "'").collect(Collectors.toList());
            sql.append(" and exam_record_status in (").append(StringUtils.join(values, ",")).append(")");
        }

        sql.append(" and id>").append(startId);

        if (shardTotal > 1) {
            sql.append(" and mod(id,").append(shardTotal).append(")=").append(shardIndex);
        }

        sql.append(" order by id asc");
        sql.append(" limit ").append(batchSize);

        log.debug(sql.toString());
        return jdbcTemplate.query(sql.toString(), new BeanPropertyRowMapper(ExamRecordDataEntity.class));
    }

}
