package cn.com.qmth.examcloud.core.oe.task.service.job;

import cn.com.qmth.examcloud.core.oe.student.dao.ExamRecordDataRepo;
import cn.com.qmth.examcloud.core.oe.student.dao.entity.ExamRecordDataEntity;
import cn.com.qmth.examcloud.core.oe.task.service.ExamRecordDataService;
import cn.com.qmth.examcloud.support.Constants;
import cn.com.qmth.examcloud.support.enums.ExamRecordStatus;
import cn.com.qmth.examcloud.support.examing.ExamRecordData;
import cn.com.qmth.examcloud.web.exception.SequenceLockException;
import cn.com.qmth.examcloud.web.helpers.SequenceLockHelper;
import org.apache.commons.collections4.CollectionUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.util.Date;
import java.util.List;

/**
 * 处理交卷后考试数据任务
 */
@Component
public class AfterHandInExamJobHandler {

    private static final Logger log = LoggerFactory.getLogger(AfterHandInExamJobHandler.class);

    @Autowired
    private ExamRecordDataRepo examRecordDataRepo;

    @Autowired
    private ExamRecordDataService examRecordDataService;

    public void run(int shardTotal, int shardIndex, String jobParam) throws Exception {
        int batchNo = 0, batchSize = 200;
        int sumTodoSize = 0;
        Long startId = 0L;
        while (true) {
            List<ExamRecordDataEntity> todoList = examRecordDataService.getExamRecordDataUnSyncList(
                    shardTotal, shardIndex, batchSize, startId, ExamRecordStatus.EXAM_HAND_IN, ExamRecordStatus.EXAM_AUTO_HAND_IN);

            batchNo++;
            sumTodoSize += todoList.size();
            log.warn("   分片任务_{}_{}_STEP2  sumTodoSize:{}  curTodoSize:{}  batchNo:{}  startId:{}",
                    shardTotal, shardIndex, sumTodoSize, todoList.size(), batchNo, startId);

            if (CollectionUtils.isEmpty(todoList)) {
                break;
            }

            int curTodoIndex = 0;
            for (ExamRecordDataEntity data : todoList) {
                final String lockKey = Constants.EXAM_CONTROL_LOCK_PREFIX + data.getStudentId();
                try {
                    SequenceLockHelper.getLockSimple(lockKey);

                    // 处理业务
                    curTodoIndex++;
                    this.handler(data.getId(), todoList.size(), curTodoIndex);
                } catch (Exception e) {
                    if (e instanceof InterruptedException) {
                        // 若线程终止，则抛出交由任务调度中心处理
                        log.warn("当前任务线程被终止！examRecordDataId:{}, error:{}", data.getId(), e.getMessage());
                        throw e;
                    } else if (e instanceof SequenceLockException) {
                        // 若锁问题，下次会继续执行
                        log.warn("当前任务获取锁失败！examRecordDataId:{}, redisKey:{}", data.getId(), lockKey);
                    } else {
                        // 若异常，下次会继续执行（需要排查原因）
                        log.error("当前任务处理失败！examRecordDataId:{}, error:{}", data.getId(), e.getMessage());
                    }
                } finally {
                    SequenceLockHelper.releaseLockSimple(lockKey);
                }
            }

            startId = todoList.get(todoList.size() - 1).getId();
            todoList.clear();
        }
    }

    /**
     * 仅处理已交卷的考试记录
     */
    private void handler(Long examRecordDataId, int curTodoSize, int curTodoIndex) throws Exception {
        ExamRecordData examRecordDataCache = examRecordDataService.getExamRecordDataCache(examRecordDataId);
        if (examRecordDataCache == null) {
            log.warn("redisKey OE_ERD:{} value not exist, Set examRecordStatus=EXAM_ERROR", examRecordDataId);
            examRecordDataRepo.updateExamRecordStatusById(ExamRecordStatus.EXAM_ERROR, new Date(), examRecordDataId);
            return;
        }

        if (ExamRecordStatus.EXAM_HAND_IN != examRecordDataCache.getExamRecordStatus()
                && ExamRecordStatus.EXAM_AUTO_HAND_IN != examRecordDataCache.getExamRecordStatus()) {
            log.warn("Mysql&Redis examRecordStatus value not equal, should EXAM_HAND_IN or EXAM_AUTO_HAND_IN but {}, examRecordDataId:{}",
                    examRecordDataCache.getExamRecordStatus(), examRecordDataId);
            // 状态出现不一致情况时，则按Redis中的值为准，下次会按实际状态执行
            examRecordDataRepo.updateExamRecordStatusById(examRecordDataCache.getExamRecordStatus(), new Date(), examRecordDataId);
            return;
        }

        // 交卷时间（手动交卷取endTime，自动交卷取cleanTime）
        Date handInTime = examRecordDataCache.getEndTime() != null ? examRecordDataCache.getEndTime() : examRecordDataCache.getCleanTime();
        long diff = (System.currentTimeMillis() - handInTime.getTime()) / 1000;
        log.warn("交卷类型:{} 交卷距离现在已{}秒, examRecordDataId:{}, studentId:{}, curTodoSize:{}, curTodoIndex:{}",
                examRecordDataCache.getHandInExamType(), diff, examRecordDataId, examRecordDataCache.getStudentId(), curTodoSize, curTodoIndex);

        try {
            // 处理人脸、活体、违纪、算分等
            long startTime = System.currentTimeMillis();
            examRecordDataService.processAfterHandInExam(examRecordDataCache);
            long cost = (System.currentTimeMillis() - startTime) / 1000;
            log.warn("processAfterHandInExam ok, examRecordDataId:{}, studentId:{}, examType:{}, cost:{}s",
                    examRecordDataCache.getId(), examRecordDataCache.getStudentId(), examRecordDataCache.getExamType(), cost);
        } catch (Exception e) {
            log.error("processAfterHandInExam fail, examRecordDataId:{}, error:{}", examRecordDataCache.getId(), e.getMessage());
            throw e;
        }
    }

}
