package cn.com.qmth.examcloud.core.oe.task.service.job;

import cn.com.qmth.examcloud.core.oe.student.api.ExamRecordDataCloudService;
import cn.com.qmth.examcloud.core.oe.student.api.request.HandInExamReq;
import cn.com.qmth.examcloud.core.oe.student.dao.ExamRecordDataRepo;
import cn.com.qmth.examcloud.core.oe.student.dao.entity.ExamRecordDataEntity;
import cn.com.qmth.examcloud.core.oe.task.service.ExamRecordDataService;
import cn.com.qmth.examcloud.core.oe.task.service.ExamingSessionService;
import cn.com.qmth.examcloud.support.Constants;
import cn.com.qmth.examcloud.support.enums.ExamRecordStatus;
import cn.com.qmth.examcloud.support.enums.HandInExamType;
import cn.com.qmth.examcloud.support.examing.ExamRecordData;
import cn.com.qmth.examcloud.support.examing.ExamingActivityTime;
import cn.com.qmth.examcloud.support.examing.ExamingSession;
import cn.com.qmth.examcloud.support.redis.RedisKeyHelper;
import cn.com.qmth.examcloud.web.exception.SequenceLockException;
import cn.com.qmth.examcloud.web.helpers.SequenceLockHelper;
import cn.com.qmth.examcloud.web.redis.RedisClient;
import org.apache.commons.collections4.CollectionUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.util.Date;
import java.util.List;

/**
 * 处理交卷前考试数据任务
 */
@Component
public class BeforeHandInExamJobHandler {

    private static final Logger log = LoggerFactory.getLogger(BeforeHandInExamJobHandler.class);

    @Autowired
    private ExamRecordDataRepo examRecordDataRepo;

    @Autowired
    private ExamRecordDataService examRecordDataService;

    @Autowired
    private ExamingSessionService examingSessionService;

    @Autowired
    private ExamRecordDataCloudService examRecordDataCloudService;

    @Autowired
    private RedisClient redisClient;

    public void run(int shardTotal, int shardIndex, String jobParam) throws Exception {
        int batchNo = 0, batchSize = 200;
        int sumTodoSize = 0;
        Long startId = 0L;
        while (true) {
            List<ExamRecordDataEntity> todoList = examRecordDataService.getExamRecordDataUnSyncList(
                    shardTotal, shardIndex, batchSize, startId, ExamRecordStatus.EXAM_ING);

            batchNo++;
            sumTodoSize += todoList.size();
            log.warn("  分片任务_{}_{}_STEP1  sumTodoSize:{}  curTodoSize:{}  batchNo:{}  startId:{}",
                    shardTotal, shardIndex, sumTodoSize, todoList.size(), batchNo, startId);

            if (CollectionUtils.isEmpty(todoList)) {
                break;
            }

            for (ExamRecordDataEntity data : todoList) {
                final String lockKey = Constants.EXAM_CONTROL_LOCK_PREFIX + data.getStudentId();
                try {
                    SequenceLockHelper.getLockSimple(lockKey);

                    // 处理业务
                    this.handler(data.getId());
                } catch (Exception e) {
                    if (e instanceof InterruptedException) {
                        // 若线程终止，则抛出交由任务调度中心处理
                        log.warn("当前任务线程被终止！examRecordDataId:{}, error:{}", data.getId(), e.getMessage());
                        throw e;
                    } else if (e instanceof SequenceLockException) {
                        // 若锁问题，下次会继续执行
                        log.warn("当前任务获取锁失败！examRecordDataId:{}, redisKey:{}", data.getId(), lockKey);
                    } else {
                        // 若异常，下次会继续执行（需要排查原因）
                        log.error("当前任务处理失败！examRecordDataId:{}, error:{}", data.getId(), e.getMessage());
                    }
                } finally {
                    SequenceLockHelper.releaseLockSimple(lockKey);
                }
            }

            startId = todoList.get(todoList.size() - 1).getId();
            todoList.clear();
        }
    }

    /**
     * 仅处理正在进行中的考试记录
     */
    private void handler(Long examRecordDataId) throws Exception {
        ExamRecordData examRecordDataCache = examRecordDataService.getExamRecordDataCache(examRecordDataId);
        if (examRecordDataCache == null) {
            log.warn("redisKey OE_ERD:{} value not exist, Set examRecordStatus=EXAM_ERROR", examRecordDataId);
            examRecordDataRepo.updateExamRecordStatusById(ExamRecordStatus.EXAM_ERROR, new Date(), examRecordDataId);
            return;
        }

        if (ExamRecordStatus.EXAM_ING != examRecordDataCache.getExamRecordStatus()) {
            log.warn("Mysql&Redis examRecordStatus value not equal, should EXAM_ING but {}, examRecordDataId:{}",
                    examRecordDataCache.getExamRecordStatus(), examRecordDataId);
            // 状态出现不一致情况时，则按Redis中的值为准，下次会按实际状态执行
            examRecordDataRepo.updateExamRecordStatusById(examRecordDataCache.getExamRecordStatus(), new Date(), examRecordDataId);
            return;
        }

        long diff = System.currentTimeMillis() - examRecordDataCache.getEnterExamTime().getTime();
        if (diff <= 180000L) {
            // 跳过刚进入考试3分钟内的记录
            log.debug("enterExamTime just 3 minutes, skip... examRecordDataId:{}", examRecordDataId);
            return;
        }

        ExamingSession examingSession = examingSessionService.getExamingSession(examRecordDataCache.getStudentId());
        if (examingSession == null) {
            // 考试会话不存在，则自动交卷
            log.warn("redisKey OE_SESSION:{} value not exist, do autoHandInExam, examRecordDataId:{}", examRecordDataCache.getStudentId(), examRecordDataId);
            this.autoHandInExam(examRecordDataCache);
            return;
        }

        // 超过断点续考时间，则自动交卷
        if (this.overBreakpointTime(examingSession)) {
            log.warn("overBreakpointTime do autoHandInExam, examRecordDataId:{}, studentId:{}", examRecordDataCache.getId(), examRecordDataCache.getStudentId());
            this.autoHandInExam(examRecordDataCache);
            return;
        }

        // 异常情况超过半天仍未交卷，则自动交卷
        if (diff >= 6L * 3600000L) {
            log.warn("enterExamTime over 6 hours... do autoHandInExam, examRecordDataId:{}, studentId:{}", examRecordDataCache.getId(), examRecordDataCache.getStudentId());
            this.autoHandInExam(examRecordDataCache);
        }
    }

    /**
     * 是否超过断点续考时间
     */
    private boolean overBreakpointTime(ExamingSession examingSession) throws Exception {
        long now = System.currentTimeMillis();

        String activeTimeKey = RedisKeyHelper.getBuilder().examingActiveTimeKey(examingSession.getExamRecordDataId());
        ExamingActivityTime activeTimeCache = redisClient.get(activeTimeKey, ExamingActivityTime.class);

        long activeTime;
        if (activeTimeCache == null) {
            log.warn("redisKey OE_ACTIVE:{} value not exist, examReconnectTime:{} minutes",
                    examingSession.getExamRecordDataId(), examingSession.getExamReconnectTime());
            activeTime = System.currentTimeMillis();
        } else {
            activeTime = activeTimeCache.getActiveTime();
        }

        return (now - activeTime) >= examingSession.getExamReconnectTime() * 60L * 1000L;
    }

    /**
     * 自动交卷
     */
    private void autoHandInExam(ExamRecordData examRecordDataCache) throws Exception {
        try {
            HandInExamReq req = new HandInExamReq();
            req.setExamRecordDataId(examRecordDataCache.getId());
            req.setHandInExamType(HandInExamType.AUTO);

            long startTime = System.currentTimeMillis();
            examRecordDataCloudService.handInExam(req);
            long cost = (System.currentTimeMillis() - startTime) / 1000;
            log.warn("autoHandInExam ok, examRecordDataId:{}, studentId:{}, examType:{}, cost:{}s",
                    examRecordDataCache.getId(), examRecordDataCache.getStudentId(), examRecordDataCache.getExamType(), cost);
        } catch (Exception e) {
            log.error("autoHandInExam fail, examRecordDataId:{}, error:{}", examRecordDataCache.getId(), e.getMessage());
            throw e;
        }
    }

}
